%**********************************************************
%   Directory listing for opening or inserting files.
%
%   Totally dependent on line format and sorting.
%   Any change to format implies revising the code.
%
%   Upon invocation, reply to the minibuffer prompt
%   is read and validated. If the entry identifies
%   one unique file for editing, this file is opened.
%   Otherwise a directory listing is displayed.
%   The list is scrollable and has its own menus.
%   Enter loads a file or opens a directory,
%   Backspace takes to upper directory.
%**********************************************************

%  _traceback = 1;


%---    required prototypes  ------------------------------

define dls_create_menu ()       ;
define dls_end ()               ;
define dls_home ()              ;
define dls_insert_file ()       ;
define dls_keymap ()            ;
define dls_read_dir (dir)       ;
define dls_upperdir ()          ;

%----------------------------------------------------------



%**********************************************************
%   Variables
%**********************************************************

private variable ShowDirs    = 0;
private variable dls_cdir    = Null_String;
private variable lMark       = NULL;
private variable nameCol     = 43;
variable InsertFromListing   = 0;


define dls (dir)
{
    %   entry point from outside, starts the listing

    sw2buf (DirListing);
    set_buffer_no_autosave ();
    dls_create_menu ();
    dls_read_dir (dir);
    flush (" ");
}


define dls_test ()
{

}


define dls_goto_nameCol()
{
    set_readonly(0);
    goto_column (nameCol);
    set_readonly(1);
}


define dls_pull_menus (keypress)
{
    if (keypress=="f")
        menu_select_menu ("DirListing.&File");

    else if (keypress=="m")
        menu_select_menu ("DirListing.&Misc");

    else if (keypress=="h")
        menu_select_menu ("DirListing.&Help");
}


define dls_jump_to_initial ()
%----------------------------------------------------------
%   Pressing a letter takes the cursor to the next entry
%   beginning with that letter. After the last entry with
%   that letter, the cursor jumps back to the first one,
%   i.e. jumping occurs in a ring. As an example,
%   pressing "a" when the cursor is resting on "t"
%   may take the cursor back in the listing.
%   Upper/lower case makes no difference.
%
%   When sub directories or links are displayed as separate
%   group, the cursor will not be restricted to a group!
%   As an example, if the cursor is on some directory,
%   you press 'q', no directory beginning with 'q' exists
%   but there are files that match. The cursor will jump
%   to these files leaving the directories.
%----------------------------------------------------------
{
    %------------------------------------------------------

    %   catch alt-<letter> for the pulldowns
    %   under ubuntu it must be <alt> on the left

    variable chk =
        LASTKEY=="f" or LASTKEY=="m" or LASTKEY=="h";

    %   LAST_CHAR delivers escape!

    if(chk)
    {
        variable s = substr (LASTKEY,2,1);
        if(strlen(s)) dls_pull_menus (s);
        return;
    }

    if(LASTKEY[0]==27 and strlen(LASTKEY)==2) return;

    %------------------------------------------------------

    current_mini_cmd = _function_name;

    variable
        found = 0,
        cf    = 0,
        initl = strlow(LASTKEY),
        INITL = strup(LASTKEY),
        rply  = Null_String,
        prmpt = "Not found. Cycle and restart from top (y/n)?";

    variable s_initial = " ";

    if (what_line < 3);
    %   do nothing

    else if (is_whiteline);
    %   do nothing

    else
    {
        push_spot();
        dls_goto_nameCol();
        s_initial = strlow(char(what_char()));
        pop_spot();
    }

    push_mark ();

    forever
    {
        eol ();
        if (eobp) break;
        found = fsearch ("  "+initl);
        ifnot(found) break;
        dls_goto_nameCol();

        %   make this case insensitive
        if (looking_at(initl)) break;
        if (looking_at(INITL)) break;
        found=0;
    }

    if (found)
    {
        pop_mark(0);
        return;
    }

    %   is target initial past initial of start position?
    if (s_initial < initl)
    {
        flush("Not found.");
        return;
    }

    ifnot(strlen(strtrim(initl)))
    {
        dls_home ();
        return;
    }

    %   nothing found below start position, cycle around
    rply = strtrim (read__mini (prmpt, " y", " y", 2));

    ifnot (strlen(rply)) return;
    if (rply != "Y" and rply !="y") return;

    bob ();

    forever
    {
        eol ();
        if (eobp) break;
        found = fsearch ("  "+initl);
        ifnot(found) break;
        dls_goto_nameCol();

        %   make this case insensitive
        if (looking_at(initl)) break;
        if (looking_at(INITL)) break;
        found=0;
    }

    if (found) pop_mark(0);
    else pop_mark(1);
}


define dls_skipup_initial ()
%----------------------------------------------------------
%   Takes to the previous entry beginning with a different
%   initial than the current initial.
%   Not case sensitive.
%----------------------------------------------------------
{
    variable s_initial = Null_String;

    if (what_line < 3) go_up (1);

    else if (is_whiteline) go_up(1);

    else
    {
        dls_goto_nameCol();
        s_initial = strlow( char (what_char()));
    }

    ifnot(strlen(s_initial)) return;

    while (s_initial == strlow(char(what_char())))
    {
        go_up (1);
        if (is_whiteline)
        {
            go_down(1);
            return;
        }
        dls_goto_nameCol();
    }

    s_initial = strlow( char (what_char()));
    while (bsearch("  " + s_initial));
}


define dls_skip_initial ()
%----------------------------------------------------------
%   Shift not pressed:  Takes down to closest entry in list
%                       beginning with a different initial
%                       than the current initial.
%                       Not case sensitive.
%   Shift pressed:      Takes up to closest entry in list
%                       beginning with a different initial
%                       than the current initial.
%                       Not case sensitive.
%----------------------------------------------------------
{
    if (SHIFTKEY_PRESSED)
    {
        dls_skipup_initial ();
        return;
    }

    variable s_initial = Null_String;

    if (what_line < 3) go_down(1);

    else if (is_whiteline) go_down(1);

    else
    {
        dls_goto_nameCol();
        s_initial = strlow( char (what_char()));
    }

    ifnot(strlen(s_initial)) return;

    while (s_initial == strlow(char(what_char())))
    {
        go_down (1);
        if (is_whiteline)
        {
            go_up(1);
            return;
        }
        dls_goto_nameCol();
    }
}

define dls_quit ()
{
    %   useless command affects exit buffer
    %sw2buf (DirListing);
    sw2buf ("*scratch*");
    delbuf (DirListing);
}


define dls_2tree ()
{
    if(ALTKEY_PRESSED)
    {
        dls_quit();
        return;
    }

    ifnot(lngth(listsubs (dls_cdir)))
    {
        flush("No tree branch below this directory.");
        return;
    }

    %   this is just to avoid losing the return buffer
    %   when switching between DirListing and Treeview
    sw2buf (TreeView);
    delbuf (DirListing);

    dtr ();
}

define dls_load (name)
{
    %   will load a file for editing
    %   argument needs full path

    %   intrinsic read_file is find_file_cmd in ledit.c
    %   however there is a non-intrinsic read_file in file.c
    %   and it is called from find_file_cmd

    %   consider file opened in buffer
    %   while deleted on disk from external application
    %   you try to open it again from dirlisting
    %   it will just switch buffer with no error
    %   giving impression loading was ok but it was not

    variable st;
    st = stat_file (name);
    if (st == NULL)
    {
        error ("Unable to find " + name +": Refresh listing!");

        %   Some files listed by ls return 0 (non existent),
        %   these are non-writable system files like /proc/kcore.
        %   Alert: non-writable system files can crash jed!
        return;
    }

    if(bufferp(path_basename(name)))
        flush("File already open, not re-loaded.");
        %   this message will be overwritten if file changed on disk
    else assoc_delete_key (bkm, path_basename(name));

    () = find_file (name);
}


define dls_shape_list ()
{
    %   entries are already sorted by group
    %   in the order dir/regular/link
    %   sort by name in each group

    variable
        o_string      = Null_String,
        perms_letters = Null_String;

    bob();

    if (bol_fsearch ("04"))
    {
        %   directories
        newline();

        %   find the names, find the colon in hours
        %   then the separating two spaces, then skip

        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        push_mark();
        eol();
        while (bol_fsearch ("04")) eol();
        sort1 ();

        %   now convert permissions from digits to letters

        while (bol_fsearch ("04"))
        {
            push_mark();
            go_right(6);
            o_string = bufsubstr();
            bufsubstr();
            perms_letters = ed_permissions(o_string);
            bol();
            loop (6) del();
            insert ("d"+perms_letters);
        }
    }

    if (bol_fsearch("10"))
    {
        %   regular files
        newline();

        %   found the names, now find the colon in hours
        %   then the separating two spaces, then skip
        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        %   cursor on name
        push_mark();
        eol();
        while (bol_fsearch ("10")) eol();
        sort1 ();

        while (bol_fsearch ("10"))
        {
            push_mark();
            go_right(6);
            o_string = bufsubstr();
            perms_letters = ed_permissions (o_string);
            bol();
            loop (6) del();
            insert ("-"+perms_letters);
        }
    }

    if (bol_fsearch ("12"))
    {
        %   symlinks
        newline();

        %   to find the names, find the colon in hours
        %   then the separating two spaces, then skip
        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        push_mark();
        eol();

        while (bol_fsearch ("12")) eol();
        sort1 ();

        %   now convert permissions from digits to letters

        perms_letters = "l---------";

        while (bol_fsearch ("12"))
        {
            loop (6) del();
            insert (perms_letters);
        }
    }

    %   give the listing a header

    bob ();
    insert (dls_cdir);

    dls_home();
    %   highlight is positioned on first non-empty line

    set_buffer_modified_flag (0);
    set_readonly (1);
}


define dls_stmode (S)
{
    variable q, stf;

    stf = lstat_file (S);
    if (stf==NULL) return Null_String;

    %   check mode, first two digits of five
    %   41   link            1 8    32
    %   33   regular file    1      32
    %   16   directory           16

    q = stf.st_mode/1000;

    if (q==16 or q==17)
    {
        %   directory, do nothing
    }

    else if (q==33)
    {
        %   regular files are handled below

        if (dls_cdir=="/proc") return Null_String;

        if (is_substr (dls_cdir, "/proc/") == 1)
            return Null_String;

        if (dls_cdir=="/dev") return Null_String;

        if (is_substr (dls_cdir, "/dev/") == 1)
            return Null_String;
    }

    else if (q==41)
    {
        %   leave only links to files, other links out

        variable stf2 = stat_file (S);

        if (stf2 == NULL)
        {
            %   redraw screen because of interfering shell messages
            call("redraw");
            return Null_String;
        }

        if(stat_is("reg",stf2.st_mode))
        {
            %   regular file
            S = readlink(S);
        }

        else return Null_String;
    }

    %   all the remaining weird files
    else return Null_String;

    variable operms;
    operms = sprintf ("%06o",stf.st_mode);

    return strcat
    (
        operms,
        sprintf ("%12S",stf.st_size),
        strftime ("  %Y-%m-%d %H:%M  ", localtime (stf.st_mtime)),
        S
    );
}


define dls_draw_list (whichdir)
{
    variable A, B, L, i;

    A = listdir (whichdir);
    B = array_map (String_Type, &dls_stmode, A);
    B = B[where(B!=Null_String)];
    B = B[array_sort (B)];
    L = lngth(B);

    bob();

    ifnot(L)
    {
        insert(whichdir);
        newline();
        newline();
        insert("    Directory empty, press any key.");
        flush ("    Directory empty, press any key.");
        newline();
        newline();
        update(1);
        flush_input(getkey());
        dls_upperdir();
        return 1;
    }

    for (i=0;i<L;i++)
    {
        newline();
        insert(B[i]);
    }

    return 0;
}


define dls_read_dir (dir)
{
    %   second parameter, if present,
    %   will force listing of subdirs

    variable
        f_mask   = Null_String,
        file,
        flags,
        s,
        tmpstr,
        old_dir;

    old_dir = dls_cdir;

    s = file_status (dir);

    if (s==2)
    {
        %   this is a directory, put it in canonical form

        tmpstr = dircat (dir, Null_String);
        %   dircat required for proper name in full
        %   the second parameter is irrelevant

        dls_cdir = path_dirname (tmpstr);
    }

    %   allow exact hit to open immediately
    else if (s==1)
    {
        %   file exists

        (dls_cdir, f_mask) =
        (path_dirname (dir), path_basename (dir));

        tmpstr = dircat (dls_cdir, f_mask);
        dls_load (tmpstr);
        return;
    }

    else
    {
        flush ("Not available or accessible.");
        delbuf (DirListing);
        return;
    }

    sw2buf (DirListing);

    (file,,,flags) = getbuf_info ();

    setbuf_info (file, dls_cdir, DirListing, flags);

    set_status_line
    ("  " + dls_cdir + "  |  %p  |  F1=Help  ", 0);

    dls_keymap ();
    set_readonly (0);
    erase_buffer ();

    ()  = chdir (dls_cdir);

    %   list is sorted by dir/regular/link
    %   fine tuning with dls_shape_list

    s = dls_draw_list (dls_cdir);
    ifnot(s) dls_shape_list();
    flush(" ");
}


define dls_refresh ()
{
    call("redraw");
    dls_read_dir (dls_cdir);
}

define dls_down ()
{
    go_down (1);

    %   must accommodate case where sub-dirs are displayed
    if(is_whiteline)
    {
        go_down(1);
        if(is_whiteline) () = bsearch("-");
    }
    bol ();
}


define dls_up ()
{
    go_up (1);

    %   must accommodate case where sub-dirs are displayed
    if(is_whiteline)
    {
        go_up(1);
        bol();
        ifnot(looking_at("d")) () = fsearch("-");
    }

    bol ();
    if(what_line < 4) recenter (window_info ('r'));
}


define dls_home ()
{
    bob();
    eol();
    () = fsearch ("-");
}


define dls_end ()
{
    eob();
    () = bsearch ("-");
}


define dls_getfile ()
{
    variable name, type, tmps;

    % (name, type) = dls_getfile ();
    % name = name of file or directory

    % type = -1 : other
    % type = 0  : file read-only
    % type = 1  : file editable
    % type = 2  : directory
    % type = 3  : link

    bol ();
    if      (looking_at ("l")) type = 3;
    else if (looking_at ("d")) type = 2;
    else if (looking_at ("-"))
    {
        push_mark();
        go_right(10);
        tmps=bufsubstr();
        type = is_substr (tmps,"w");
        if (type) type=1;
    }
    else return (Null_String, -1);

    dls_goto_nameCol();
    push_mark ();
    eol ();
    name = bufsubstr ();

    return (name, type);
}


define dls_menubar ()
{
    select_pulldown_file();
}


define dls_help ()
{
    select_pulldown_help ();
}


define dls_display_help ()
{
    variable
        stringA="DirListing",
        stringB="press F1",
        stringC="&Quick Help: Dir Listing";

    %   display_hlp2 bad idea
    display_help (stringA,stringB,stringC);
    dls (dls_cdir);
    flush(" ");
}


define dls_position (name)
%   used for links to files
%   link is not opened, cursor is positioned on it
{
    variable
        found,
        p_name = path_dirname (name),
        f_name = path_basename (name);

    dls_read_dir (p_name);
    bob ();

    forever
    {
        found = re_fsearch (f_name);

        ifnot (found)
        {
            bob ();
            flush("Not found.");
            break;
        }

        if (nameCol == what_column ()) break;
        eol ();
    }
}


define dls_open_current ()
%----------------------------------------------------------
%   Will open a file from DirListing in own buffer.
%
%   However, if Insert was requested, the current file
%   will be inserted into the requesting buffer.
%----------------------------------------------------------
{
    %   called when pressing enter on a name in DirListing

    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;
    (name, type) = dls_getfile ();

    if (bufferp(InsertBuffer) and (type==1 or type==0))
    {
        variable
            yn = Null_String,
            i_prmpt = "Insert into buffer " + InsertBuffer + " (y/n)?";

        yn = strlow(strtrim(read__mini (i_prmpt, "n", "y", 2)));

        if (yn=="y")
        {
            dls_insert_file ();
            InsertBuffer = Null_String;
            return;
        }
        flush("No insertion.");
        sleep(1);
    }

    InsertBuffer = Null_String;

    name = dircat (dls_cdir, name);

    if (type==0 or type==1)
    {
        %   1 is a file for editing
        %   0 is read-only

        delbuf(whatbuf);
        dls_load (name);
        %   loading will close dir listing
    }

    else if (type==3)
        flush("Reference only, links are not opened.");

    %   this should be a directory or else trouble
    else
    {
        variable A = listdir(name);
        variable B = listsubs(name);

        if(lngth(B))
            dls_read_dir (name,name);

        else if(lngth(A))
            dls_read_dir (name,name);

        else flush ("Directory empty.");
    }
}


define dls_copy_path ()
%----------------------------------------------------------
%   Copy path and name of current file to clipboard.
%   Directories: path will also be copied.
%   Links: target path will be copied
%----------------------------------------------------------
{
    variable name, type, thisbf;

    (name, type) = dls_getfile ();

    if (type < 0)
    {
        flush ("No.");
        return;
    }

    name = dircat (dls_cdir, name);

    %   useful quirk when current entry is link:
    %   no catenation since link starts with a slash

    thisbf = whatbuf();
    sw2buf("*scratch*");
    eob();
    newline ();
    push_mark ();
    insert(name);
    call("kill_region");
    sw2buf(thisbf);
    flush("Full path copied to clipboard.");
}


define dls_copy_file()
%----------------------------------------------------------
%   Will copy a file to clipboard so it can be pasted
%   elsewhere
%----------------------------------------------------------
{
    variable name, type, thisbf, strct;

    (name, type) = dls_getfile ();

    if (type!=0 and type!=1)
    {
        flush ("No.");
        return;
    }

    name = dircat (dls_cdir, name);

    strct = stat_file (name);
    if(strct.st_size==0)
    {
        flush("File is empty.");
        return;
    }

    thisbf = whatbuf();
    sw2buf("*scratch*");
    eob();
    newline ();
    push_mark ();
    newline();
    () = insert_file(name);
    eob();
    newline();
    call("kill_region");
    sw2buf(thisbf);
    flush("File copied to clipboard.");
}


define dls_insert_file ()
%----------------------------------------------------------
%   will insert file from DirListing into edit buffer
%   after command from DirListing menu
%   request must originate from edit buffer
%----------------------------------------------------------
{
    %   if current entry not a file, do nothing, say no
    %   if current entry is a file,
    %       insertbuffer empty  refuse
    %       else                insert

    variable name, type;
    (name, type) = dls_getfile ();

    if (type!=1 and type!=0)
    {
        verror ("No.");
        return;
    }

    ifnot (bufferp(InsertBuffer))
    {
        verror ("Don't know where to insert.");
        return;
    }

    else
    {
        delbuf(whatbuf);
        sw2buf(InsertBuffer);
    }

    InsertBuffer=Null_String;
    name = dircat (dls_cdir, name);

    eol ();
    newline();
    () = insert_file (name);
    InsertBuffer = Null_String;
}


define dls_upperdir ()
{
    if (dls_cdir==slash)
    {
        () = chdir (slash);
        dtr();
        return;
    }

    () = chdir ("..");
    dls_read_dir (".");
}


define dls_pgup ()
%----------------------------------------------------------
%   will move up cursor one page
%----------------------------------------------------------
{
    call ("page_up");
    bol();
    ifnot(ffind("-")) () = fsearch("-");
    bol();
}


define dls_pgdn ()
%----------------------------------------------------------
%   will move down cursor one page
%----------------------------------------------------------
{
    call ("page_down");
    bol();
    ifnot(ffind("-")) () = bsearch("-");
    bol();
}


define dls_save ()
{
    variable L, b_name = Null_String;

    b_name = read_path_from_mini ("Save listing to file:");

    %   aborted
    if (LAST_CHAR != 13) return;

    %   if response empty, do nothing
    L = strlen(b_name);
    ifnot(L) return;
    if(b_name[L-1] == '/') return;

    %   select entire listing for saving

    push_spot ();
    bob ();
    push_mark ();
    eob ();

    %   check file status before proceeding

    variable b_status = file_status (b_name);

    if (b_status == 2)
        error ("Name already in use for a directory");

    else if (b_status == -1)
        error ("File exists and is not accessible");

    else if (b_status == -2)
        error ("Invalid path");

    else if (b_status == -3)
        error ("Undetermined error");

    else if (b_status == 0)
    {
        write_region_to_file (b_name);
    }

    else if (b_status == 1)
    {
        variable ch = get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch=='a' or ch=='A')
            append_region_to_file (b_name);

        else if (ch=='o' or ch=='O')
            write_region_to_file (b_name);
    }
    pop_spot ();
    dls_refresh (dls_cdir);
}


define dls_delete_file ()
{
    %   called when pressing Del on a name in DirListing

    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;

    (name, type) = dls_getfile ();

    %   take note of following file to position cursor after deleting
    go_down(1);
    variable name_nxt=Null_String;
    ifnot (is_whiteline) (name_nxt,) = dls_getfile ();
    go_up(1);

    name = dircat (dls_cdir, name);

    if (type==1 or type==0)
    {
        %   this is a file for editing

        variable ch =
        get_mini_response ("Delete current file? (y/n)  ");

        if (ch!='y' and ch!='Y') return;

        delete_file(name);
        dls_refresh ();

        if(strlen(name_nxt))
            ()=fsearch("  "+name_nxt);
        else eob();
        while (is_whiteline ()) go_up (1);
    }

    else flush ("No.");
}


define dls_grep (s_strng)
{
    if (_NARGS == 0)
        s_strng = Null_String;

    variable
        sstr    = Null_String,
        fmask   = Null_String,
        fopts   = Null_String,
        cmdstr  = Null_String,
        result  = Null_String,
        rfile   = "!_grep_results.tmp";

    current_mini_cmd = _function_name;

    if(strlen(s_strng)) sstr=s_strng;
    else
        sstr = read__mini ("Find what:", s_strng, s_strng, 0);

    ifnot (strlen(sstr)) return;

    %   make sure search string is quoted
    sstr = "\""+ sstr + "\"";

    fmask = read__mini ("Files to search:", "*", "*", 0);
    %   no file completion here

    ifnot (strlen(strtrim(fmask))) return;

    fmask = "\""+ fmask + "\"";
    fmask = strcat(" * --include=",fmask);

    fopts = read__mini ("Options to use:", "in", "in", 2);

    ifnot(strlen(fopts)) return;

    fopts = "-s" + fopts;

    () = delete_file (rfile);
    sw2buf(rfile);
    erase_buffer();

    cmdstr = "grep " + fopts + " " + sstr + " " + fmask;

    flush ("Grepping ... To stop press <ctrl><backslash>.");

    try
    {
        result = run_shell_cmd (cmdstr);
    }
    catch AnyError:
    {
        %   slang errors, not grep errors
        delbuf(whatbuf);
        dls_refresh();
        flush("Stopped, error.");
        return;
    }

    %   as long as grep is running,
    %   control will be in the minibuffer
    %   actually in a separate process

    ifnot (MINIBUFFER_ACTIVE)
    {
        if (result==1)
        {
            delbuf (rfile);
            dls (".");
            flush("Not found. Executed: " + cmdstr);

            %%  add to stack, sometimes underflow occurs
            string(cmdstr);
            delbuf(whatbuf);
            return;
        }

        else
        {
            %   result 0    everything ok
            %   result 2    partial results, e.g. inaccessible files

            bob();
            insert(dls_cdir+"/");
            newline();
            newline();
            insert("  <alt><kp_enter>");
            newline();
            insert
            (
            "  Press this key on a line to jump to the file/line referenced."
            );
            newline();
            insert
            ("  If not found, prefix the path above.");
            loop(3) newline();

            %   save buffer to disk
            () = write_buffer (rfile);

            bob();
            replace (char(9), " ");
            bob();
            flush("Executed: " + cmdstr);

            if(bobp==eobp)
            {
                delbuf (whatbuf);
                delete_file (rfile);
                write_scratch (cmdstr);
                dls (dls_cdir);
                flush ("Nothing found for: " + cmdstr);
                return;
            }
            if(result > 1)
            {
                call("redraw");
                bob();
                newline();
                insert("Partial results, some files not accessed.");
                newline();
                insert("Prefix a backslash to regex operators if they are meant to be literals.");
                newline();
                newline();
                bob();
            }
        }
    }
}


define dls_list_keys ()
{
    set_readonly (0);
    erase_buffer ();
    bob ();
    $1 = "dirlist_slimkmap";

    sw2buf ($1);
    newline();
    dump_bindings ($1);

    %   suppress lines containing self-insert characters
    bob ();
    while (fsearch ("dls_jump_to_initial")) delete_line ();

    readable_keys();

    bob ();
    replace     ("dtr", "Tree View");

    bob ();
    replace     ("^M   ",  "Enter");

    bob ();
    push_mark ();
    eob();
    go_left(1);
    sort2 ();
    flush (" ");
    %   this sort is not case sensitive

    set_buffer_modified_flag (0);
    variable m = "&Keybindings DirListing";
    dummy_menu(m);
    set_status_line("  Press any key to exit",0);
    hide_status_indicators(1);
    eob();
    loop(8) newline();
    bob ();
    while(is_whiteline) go_down(1);
    update (1);

    flush_input(getkey);
    delbuf(whatbuf);
    dls (dls_cdir);
}


define dls_cleanup ()
{
    %   get rid of bakups and tmp files
    () = run_shell_cmd ("rm *~; rm *bak; rm *tmp");
    () = run_shell_cmd ("rm -f \.?*~");
    dls_refresh ();
}


define dls_shellcmd ()
%----------------------------------------------------------
%   Execute a non-interactive shell command from DirListing
%   keeping cursor approx. on same screen line
%   sending output to a file in current directory
%----------------------------------------------------------
{
    variable
        exit    = 0,
        sstr    = Null_String,
        rfile   = "!_shellcmd_results.tmp",
        type    = 0,
        name    = Null_String;

    (name, type) = dls_getfile ();
    if (type==2) name = dircat (dls_cdir, name);

    current_mini_cmd = _function_name;
    sstr = read__mini ("*Non-interactive* Shell Command:", name, name, 2);

    ifnot (strlen(sstr)) return;

    () = delete_file (rfile);
    sw2buf(rfile);
    erase_buffer();

    try
    {
        exit = run_shell_cmd (sstr);
    }
    catch AnyError:
    {
        flush("Error running shell command.");
        delbuf(whatbuf);
        return;
    }

    %   as long as the shell command is running,
    %   control will be in the minibuffer

    ifnot (MINIBUFFER_ACTIVE)
    {
        variable ln = 0;
        %   save buffer to disk if any messages
        bob();
        if (bobp != eobp)
            ln = write_buffer (rfile);

        delbuf (rfile);
        dls_refresh ();

        if(ln) fsearch(rfile);
        flush("Exit " + string(exit) + " from: " + sstr);
    }
}

define dls_recent ()
{
    recent_dirs();
}

define dls_create_menu ()
{
    %   The menu bar is created only once.
    %   When the create command is issued again,
    %   there is a check for the existence of the menu bar.
    %   To avoid garbage, delete before appending.

    menu_create_menu_bar (DirListing);
    menu_use_menu_bar (DirListing);

    menu_append_popup (DirListing, "&File");
    menu_append_popup (DirListing, "&Misc");
    menu_append_popup (DirListing, "&Help");

    $1 = DirListing+".&File";

    menu_delete_items ($1);
    menu_append_item ($1, "&Open..",            "dls_open_current");
    menu_append_item ($1, "D&elete",            "dls_delete_file");
    menu_append_separator ($1);
    menu_append_item ($1, "&Close Listing",     "dls_quit");
    menu_append_item ($1, "&Save Listing..",    "dls_save");
    menu_append_item ($1, "&Refresh Listing",   "dls_refresh");
    menu_append_separator ($1);
    menu_append_item ($1, "&Upper Directory",   "dls_upperdir");
    menu_append_item ($1, "Recent &Dirs..",     "dls_recent");
    menu_append_item ($1, "&Tree..",            "dls_2tree");

    $1 = DirListing+".&Misc";

    menu_delete_items ($1);
    menu_append_item ($1, "Copy F&ile",         "dls_copy_file");
    menu_append_item ($1, "Copy &Path",         "dls_copy_path");
    menu_append_separator ($1);
    menu_append_item ($1, "&Find in Files..",   "dls_grep");
    menu_append_item ($1, "&Shell Command..",   "dls_shellcmd");
    menu_append_item ($1, "&CleanUp Directory", "dls_cleanup");

    $1 = DirListing+".&Help";

    menu_delete_items ($1);
    menu_append_item ($1, "&Quick Help",        "dls_display_help");
    menu_append_item ($1, "&List Keys",         "dls_list_keys");
}


define dls_keymap ()
{
    reset_slim_keymap ();
    $1 = "dirlist_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    %   jump-by-letter-press
    %   e.g. m to go the the next initial m
    %   allow upper ASCII characters
    variable i;
    for (i=32; i<256; i++)
        definekey ("dls_jump_to_initial", char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);

    definekey ("dls_copy_path",     "^C",       $1);
    definekey ("dls_shellcmd",      "^E",       $1);
    definekey ("dls_grep",          "^F",       $1);
    definekey ("dls_refresh",       "^R",       $1);
    definekey ("dls_save",          "^S",       $1);
    definekey ("dls_test",          "^T",       $1);
    definekey ("dls_quit",          "^W",       $1);
    definekey ("dls_upperdir",      Key_BS,     $1);
    definekey ("dls_delete_file",   Key_Del,    $1);
    definekey ("dls_down",          Key_Down,   $1);
    definekey ("dls_end",           Key_End,    $1);
    definekey ("dls_open_current",  Key_Enter,  $1);
    definekey ("dls_quit",          Key_Escape, $1);
    definekey ("dls_help",          Key_F1,     $1);
    definekey ("dls_menubar",       Key_F10,    $1);
    definekey ("dls_2tree",         Key_F4,     $1);
    definekey ("dls_home",          Key_Home,   $1);
    definekey ("dls_pgdn",          Key_PgDn,   $1);
    definekey ("dls_pgup",          Key_PgUp,   $1);
    definekey ("dls_skip_initial",  Key_Tab,    $1);
    definekey ("dls_up",            Key_Up,     $1);
}
